package outlab1;

import java.util.ArrayList;
import java.util.Collections;
import edu.princeton.cs.algs4.In;

/**
 * @author Ben Miller
 * @author Lauren Helbling
 */

public class JobScheduler {

    private int globalTime = 0;
    private ArrayList<Job> notQueuedJobs = new ArrayList<>();
    private ArrayList<Job> queuedJobs = new ArrayList<>();
    private ArrayList<Job> finishedJobs = new ArrayList<>();
    private int numProcessors;

    /**
     * Instantiates a JobScheduler. It takes an In object, reads data from it, places jobs in a queue and sorts them by arrival time.
     * @param in
     * @param numProcessors
     */
    public JobScheduler(In in, int numProcessors){
        this.numProcessors = numProcessors;

        while(in.hasNextLine()){
            
            try {
                Job a = new Job(in.readInt(), in.readInt(), in.readInt(), in.readInt());
                notQueuedJobs.add(a);
                
            } catch (Exception e) {
            }
        }
        if (notQueuedJobs.isEmpty()) System.out.println("No jobs found");
        Collections.sort(notQueuedJobs);
    }

    /**
     * checks if JobScheduler is completely done
     * @return
     */
    public boolean isFinished(){
        return (notQueuedJobs.isEmpty() && queuedJobs.isEmpty());
    }

    /**
     * Progresses the job scheduler 1 second.
     */
    public void work(){
        System.out.print("Time " + String.format("%-4d", this.globalTime) + ": "); //initial time print
        for (int i = 0; i < notQueuedJobs.size(); i++) { //Adds jobs to queue once they've arrived
            if (notQueuedJobs.get(i).arrivalTime <= this.globalTime) {
                this.addToQueue(notQueuedJobs.remove(i--));
            }else{
                break;
            }
        }
        if (this.queuedJobs.isEmpty()) System.out.println();
        this.workProcess(); //works number of processes jobs in front of queue

        for (Job a: queuedJobs) { //runs inQueue on every job in queue
            a.inQueue();
        }
        this.globalTime++;

    }

    /**
     * prints message for each job and works it for one second
     */
    public void workProcess(){ 
        boolean printOnce = false;
        String message;
        if (queuedJobs.size() < numProcessors) {
            for (int i = 0; i < queuedJobs.size(); i++) {
                message = queuedJobs.get(i).workOn();
                if (!printOnce) {
                    System.out.println(message);
                    printOnce = true;
                } else {
                    System.out.println("           " + message);
                }
                if (queuedJobs.get(i).getDuration() <= 0) {
                    queuedJobs.get(i).inQueue();
                    queuedJobs.get(i).jobFinishedMessage();
                    finishedJobs.add(queuedJobs.get(i));
                    queuedJobs.remove(i--);
                }
            }
        } else {
            int temp = numProcessors;
            for (int i = 0; i < numProcessors; i++) {
                message = queuedJobs.get(i).workOn();
                if (!printOnce) {
                    System.out.println(message);
                    printOnce = true;
                } else {
                    System.out.println("           " + message);
                }
                if (queuedJobs.get(i).getDuration() <= 0) {
                    queuedJobs.get(i).inQueue();
                    queuedJobs.get(i).jobFinishedMessage();
                    finishedJobs.add(queuedJobs.get(i));
                    queuedJobs.remove(i--);
                    numProcessors--;
                }
            }
            numProcessors = temp;

        }

    }

    /**
     * Adds given job to queue in its proper priority. It inserts jobs in front of first job with lower priority
     * @param job
     */
    public void addToQueue(Job job){ 
        for (int i = 0; i < queuedJobs.size(); i++){
            if (job.priority > queuedJobs.get(i).priority){
                this.queuedJobs.add(i, job);
                return;
            }
        }
        this.queuedJobs.add(job);
    }


    /**
     * prints details for all finished jobs
     */
    public void printFinished(){ 
        int i = 1;
        for (Job a : this.finishedJobs){
            System.out.print(String.valueOf(i++));
            a.jobFinishedMessage();
        }

    }





    /**
     * Job class for use in job scheduler
     */
    public class Job implements Comparable<Job>{
        private int idNumber;
        private int priority;
        private int arrivalTime;
        private int duration;
        private int waitingTime = 0;
        private int executionTime = 0;
        private boolean startedOn = false;

        public Job(int number, int priority, int arrivalTime, int duration) {
            this.idNumber = number;
            this.priority = priority;
            this.arrivalTime = arrivalTime;
            this.duration = duration;
        }

        /**
         * decrements duration and returns details of job
         * @return
         */
        public String workOn(){
            startedOn = true;
            duration--;
            return "Job " + this.idNumber + ": priority " + this.priority + ": and duration left " + this.duration;
        }

        /**
         * increments waiting time if it hasn't been worked on, else increments execution time
         */
        public void inQueue(){
            if (!startedOn) waitingTime++;
            else executionTime++;

        }

        public int getDuration() {
            return duration;
        }


        public int getPriority(){
            return this.priority;
        }

        public int getArrivalTime() {
            return arrivalTime;
        }

        /**
         * used for sorting jobs by arrival time
         */
        @Override
        public int compareTo(Job comparesToJob) {
            int compareArrival = comparesToJob.getArrivalTime();

            return this.arrivalTime - compareArrival;
        }

        public void jobFinishedMessage(){
            System.out.println("    Job " + this.idNumber + " is finished with Waiting Time \"" + this.waitingTime + "\" and Execution Time \"" + this.executionTime + "\"");
        }
    }

    public static void main(String[] args) {
        if (args.length > 0){
            In in = new In(args[0]);
            int numProcessors = 1;
            if (args.length > 1) numProcessors = Integer.parseInt(args[1]); //If numProcesses is inputed, sets numProcesses
            JobScheduler mainJob = new JobScheduler(in, numProcessors);

            while (!mainJob.isFinished()) mainJob.work();

            System.out.println();
            System.out.println("All Jobs Finished!");
            System.out.println();
            
            mainJob.printFinished();

        } else {
            System.out.println("No arguments passed in");
        }
    }
}
